@@ -1,6 +1,6 @@
 import asyncio
 from pathlib import Path
-from typing import Any, Dict, List, Optional, Type, TypeVar, Union
+from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union
 
 import backoff
 import httpx
@@ -18,8 +18,8 @@ from .models import (
     Configurations,
     ConfigurationsResponse,
     CVEAllowlist,
+    GCHistory,
     GeneralInfo,
-    HarborVulnerabilityReport,
     InternalConfigurationsResponse,
     IsDefault,
     Label,
@@ -55,11 +55,14 @@ from .models import (
     SystemInfo,
     Tag,
     UserCreationReq,
+    UserGroup,
+    UserGroupSearchItem,
     UserProfile,
     UserResp,
     UserSearchRespItem,
     UserSysAdminFlag,
 )
+from .models.scanner import HarborVulnerabilityReport
 from .types import JSONType
 from .utils import (
     get_artifact_path,
@@ -134,7 +137,13 @@ class HarborAsyncClient:
             self.credentials = credentials
         elif credentials_file:
             crfile = load_harbor_auth_file(credentials_file)
-            self.credentials = get_credentials(crfile.name, crfile.secret)  # type: ignore # load_harbor_auth_file guarantees these are not None
+            # TODO: perform this check somewhere else?
+            #       it's likely we will ALWAYS require a username and secret
+            if not crfile.name:
+                raise ValueError("Credentials file missing value for 'name' field")
+            elif not crfile.secret:
+                raise ValueError("Credentials file missing value for 'secret' field")
+            self.credentials = get_credentials(crfile.name, crfile.secret)
         else:
             raise ValueError(
                 "Must provide username and secret, credentials, or credentials_file"
@@ -336,7 +345,9 @@ class HarborAsyncClient:
         return urldecode_header(resp, "Location")
 
     # GET /users
-    async def get_users(self, sort: Optional[str] = None, **kwargs) -> List[UserResp]:
+    async def get_users(
+        self, sort: Optional[str] = None, **kwargs: Any
+    ) -> List[UserResp]:
         """Get all users.
 
         Parameters
@@ -398,7 +409,154 @@ class HarborAsyncClient:
         """
         await self.delete(f"/users/{user_id}", missing_ok=missing_ok)
 
-    # CATEGORY: gc
+    # CATEGORY: gc (Garbage Collection)
+
+    # GET /system/gc/schedule
+    # Get gc's schedule.
+    async def get_gc_schedule(self) -> Schedule:
+        """Get Garbage Collection schedule.
+
+        Returns
+        -------
+        Schedule
+            The gc's schedule.
+        """
+        resp = await self.get("/system/gc/schedule")
+        return construct_model(Schedule, resp)
+
+    # POST /system/gc/schedule
+    # Create a gc schedule.
+    async def create_gc_schedule(self, schedule: Schedule) -> str:
+        """Create a Garbage Collection schedule.
+
+        Parameters
+        ----------
+        schedule : Schedule
+            The schedule to create
+
+        Returns
+        -------
+        str
+            The location of the created schedule.
+        """
+        resp = await self.post("/system/gc/schedule", json=schedule)
+        return urldecode_header(resp, "Location")
+
+    # PUT /system/gc/schedule
+    # Update gc's schedule.
+    async def update_gc_schedule(self, schedule: Schedule) -> None:
+        """Update the Garbage Collection schedule.
+
+        Parameters
+        ----------
+        schedule : Schedule
+            The new schedule to set
+        """
+        await self.put("/system/gc/schedule", json=schedule)
+
+    # GET /system/gc
+    # Get gc results.
+    async def get_gc_jobs(
+        self,
+        query: Optional[str] = None,
+        sort: Optional[str] = None,
+        page: int = 1,
+        page_size: int = 10,
+    ) -> List[GCHistory]:
+        # TODO: limit number of results?
+        """Get Garbage Collection history for all jobs, optionally filtered by query.
+
+        Parameters
+        ----------
+        query : Optional[str]
+            A query string to filter the Garbage Collection results logs.
+
+            Supported query patterns are:
+
+                * exact match(`"k=v"`)
+                * fuzzy match(`"k=~v"`)
+                * range(`"k=[min~max]"`)
+                * list with union releationship(`"k={v1 v2 v3}"`)
+                * list with intersection relationship(`"k=(v1 v2 v3)"`).
+
+            The value of range and list can be:
+
+                * string(enclosed by `"` or `'`)
+                * integer
+                * time(in format `"2020-04-09 02:36:00"`)
+
+            All of these query patterns should be put in the query string
+            and separated by `","`. e.g. `"k1=v1,k2=~v2,k3=[min~max]"`
+        sort : Optional[str]
+            The sort order of the logs.
+        page : int
+            The page of results to return
+        page_size : int
+            The number of results to return per page
+
+        Returns
+        -------
+        List[GCHistory]
+            List of Garbage Collection logs.
+        """
+        # TODO: refactor this and use with every method that uses queries + pagination
+        params = {
+            "q": query,
+            "sort": sort,
+            "page": page,
+            "page_size": page_size,
+        }
+        params = {k: v for k, v in params.items() if v is not None}
+        resp = await self.get("/system/gc")
+        return [construct_model(GCHistory, g) for g in resp]
+
+    # GET /system/gc/{gc_id}/log
+    # Get gc job log.
+    async def get_gc_log(
+        self, gc_id: int, as_list: bool = True
+    ) -> Union[List[str], str]:
+        """Get log output for a specific Garbage Collection job.
+
+        Results are returned as a list of lines, or as a single string if
+        `as_list` is False.
+
+        Parameters
+        ----------
+        gc_id : int
+            The ID of the Garbage Collection job to get the log for
+        as_list : bool
+            If `True`, return the log as a list of lines, otherwise as single string.
+
+        Returns
+        -------
+        Union[List[str], str]
+            The log output for the Garbage Collection job.
+        """
+        resp = await self.get_text(f"/system/gc/{gc_id}/log")
+        if as_list:
+            return resp.splitlines()
+        return resp
+
+    # GET /system/gc/{gc_id}
+    # Get gc status.
+    async def get_gc_job(
+        self,
+        gc_id: int,
+    ) -> GCHistory:
+        """Get a specific Garbage Collection job.
+
+        Parameters
+        ----------
+        gc_id : int
+            The ID of the Garbage Collection job to get information about.
+
+        Returns
+        -------
+        GCHistory
+            Information about the Garbage Collection job.
+        """
+        resp = await self.get(f"/system/gc/{gc_id}")
+        return construct_model(GCHistory, resp)
 
     # CATEGORY: scanAll
 
@@ -511,8 +669,167 @@ class HarborAsyncClient:
         return construct_model(ConfigurationsResponse, resp)
 
     # CATEGORY: usergroup
+    # GET /usergroups/search
+    # Search groups by groupname
+    async def search_usergroups(
+        self, group_name: str, page: int = 1, page_size: int = 10
+    ) -> List[UserGroupSearchItem]:
+        """Search for user groups by group name.
+
+        Parameters
+        ----------
+        group_name : str
+            The group name to search for.
+        page : int
+            The page of results to return
+        page_size : int
+            The number of results to return per page
+
+        Returns
+        -------
+        List[UserGroupSearchItem]
+            List of user groups.
+        """
+        params = {
+            "groupname": group_name,
+            "page": page,
+            "page_size": page_size,
+        }
+        params = {k: v for k, v in params.items() if v is not None}
+        resp = await self.get("/usergroups/search", params=params)
+        return [construct_model(UserGroupSearchItem, g) for g in resp]
+
+    # POST /usergroups
+    # Create user group
+    async def create_usergroup(self, usergroup: UserGroup) -> str:
+        """Create a new user group. Returns location of the created user group.
+
+        Parameters
+        ----------
+        usergroup : UserGroup
+            The user group to create
+
+        Returns
+        -------
+        str
+            The location of the created user group
+        """
+        resp = await self.post("/usergroups", json=usergroup)
+        return urldecode_header(resp, "Location")
+
+    # GET /usergroups
+    # Get all user groups information
+    async def get_usergroups(
+        self, ldap_group_dn: Optional[str] = None, page: int = 1, page_size: int = 10
+    ) -> List[UserGroup]:
+        """Get all user groups.
+
+        Parameters
+        ----------
+        ldap_group_dn : Optional[str]
+            The LDAP group DN to filter by.
+        page : int
+            The page of results to return
+        page_size : int
+            The number of results to return per page
+
+        Returns
+        -------
+        List[UserGroup]
+            List of user groups.
+        """
+        params = {
+            "ldap_group_dn": ldap_group_dn,
+            "page": page,
+            "page_size": page_size,
+        }
+        params = {k: v for k, v in params.items() if v is not None}
+        resp = await self.get("/usergroups", params=params)
+        return [construct_model(UserGroup, g) for g in resp]
+
+    # PUT /usergroups/{group_id}
+    # Update group information
+    async def update_usergroup(self, group_id: int, usergroup: UserGroup) -> None:
+        """Update group information.
+
+        Parameters
+        ----------
+        group_id : int
+            The group ID to update.
+        usergroup : UserGroup
+            The new definition for the usergroup.
+        """
+        await self.put(f"/usergroups/{group_id}", json=usergroup)
+
+    # GET /usergroups/{group_id}
+    # Get user group information
+    async def get_usergroup(self, group_id: int) -> UserGroup:
+        """Get a user group by ID.
+
+        Parameters
+        ----------
+        group_id : int
+            The group ID to get.
+
+        Returns
+        -------
+        UserGroup
+            The user group.
+        """
+        resp = await self.get(f"/usergroups/{group_id}")
+        return construct_model(UserGroup, resp)
+
+    # DELETE /usergroups/{group_id}
+    # Delete user group
+    async def delete_usergroup(self, group_id: int, missing_ok: bool = True) -> None:
+        """Delete a user group.
+
+        Parameters
+        ----------
+        group_id : int
+            The group ID to delete.
+        missing_ok : bool
+            If `True`, do not raise an error if the group does not exist.
+        """
+        await self.delete(f"/usergroups/{group_id}", missing_ok=missing_ok)
+
     # CATEGORY: preheat
+
     # CATEGORY: replication
+
+    # PUT /replication/executions/{id}
+    # Stop the specific replication execution
+
+    # GET /replication/executions/{id}
+    # Get the specific replication execution
+
+    # GET /replication/executions/{id}/tasks
+    # List replication tasks for a specific execution
+
+    # POST /replication/policies
+    # Create a replication policy
+
+    # GET /replication/policies
+    # List replication policies
+
+    # POST /replication/executions
+    # Start one replication execution
+
+    # GET /replication/executions
+    # List replication executions
+
+    # PUT /replication/policies/{id}
+    # Update the replication policy
+
+    # GET /replication/policies/{id}
+    # Get the specific replication policy
+
+    # DELETE /replication/policies/{id}
+    # Delete the specific replication policy
+
+    # GET /replication/executions/{id}/tasks/{task_id}/log
+    # Get the log of the specific replication task
+
     # CATEGORY: label
     # CATEGORY: robot
 
@@ -602,7 +919,7 @@ class HarborAsyncClient:
         params = {
             "page": page,
             "page_size": page_size,
-            "query": query,
+            "q": query,
             "sort": sort,
         }
         params = {k: v for k, v in params.items() if v is not None}
@@ -769,7 +1086,7 @@ class HarborAsyncClient:
             otherwise, retrieve only the number of resources specified by `page_size`.
         """
         params = {
-            "query": query,
+            "q": query,
             "sort": sort,
             "page": page,
             "page_size": page_size,
@@ -859,7 +1176,7 @@ class HarborAsyncClient:
             otherwise, retrieve only the number of resources specified by `page_size`.
         """
         params = {
-            "query": query,
+            "q": query,
             "sort": sort,
             "name": name,
             "public": public,
@@ -974,7 +1291,7 @@ class HarborAsyncClient:
         """
         headers = get_project_headers(project_name_or_id)
         params = {
-            "query": query,
+            "q": query,
             "sort": sort,
             "page": page,
             "page_size": page_size,
@@ -1280,7 +1597,7 @@ class HarborAsyncClient:
             A list of Registry objects.
         """
         params = {
-            "query": query,
+            "q": query,
             "sort": sort,
             "page": page,
             "page_size": page_size,
@@ -2300,7 +2617,7 @@ class HarborAsyncClient:
             A list of repositories matching the query.
         """
         params = {
-            "query": query,
+            "q": query,
             "sort": sort,
             "page": page,
             "page_size": page_size,
@@ -2432,7 +2749,7 @@ class HarborAsyncClient:
             The list of audit logs.
         """
         params = {
-            "query": query,
+            "q": query,
             "sort": sort,
             "page": page,
             "page_size": page_size,
@@ -2459,13 +2776,49 @@ class HarborAsyncClient:
         follow_links: bool = True,
         **kwargs: Any,
     ) -> JSONType:
-        return await self._get(
+        j, next_url = await self._get(
             path,
             params=params,
             headers=headers,
             follow_links=follow_links,
             **kwargs,
         )
+        if not next_url:  # no pagination
+            return j
+
+        # Make sure j is a list
+        if not isinstance(j, list):
+            logger.warning(
+                "Unable to handle paginated results: Expected a list from 'GET %s', but got %s",
+                path,
+                type(j),
+            )
+            # TODO: add toggle for this coercion (coerce or throw exception)
+            #       or should we even accomodate this use-case? Always throw exception?
+            logger.info("Coercing value from %s to list", path)
+            j = [j]
+
+        # Send requests as long as we get next links
+        while next_url:
+            paginated, next_url = await self._get(
+                next_url,
+                # don't pass params (they should be in next URL)
+                headers=headers,
+                follow_links=follow_links,
+                **kwargs,
+            )
+            if not isinstance(paginated, list):
+                logger.warning(
+                    "Unable to handle paginated results: Expected a list from 'GET %s', but got %s",
+                    next_url,
+                    type(paginated),
+                )
+                # NOTE: we could also abort here, so we don't get partial results
+                #       but right now it's unclear whether this can ever happen
+                continue
+            j.extend(paginated)
+
+        return j
 
     @backoff.on_exception(backoff.expo, RequestError, max_time=30)
     async def get_text(
@@ -2478,7 +2831,8 @@ class HarborAsyncClient:
         """Bad workaround in order to have a cleaner API for text/plain responses."""
         headers = headers or {}
         headers.update({"Accept": "text/plain"})
-        resp = await self._get(path, params=params, headers=headers, **kwargs)
+        resp, _ = await self._get(path, params=params, headers=headers, **kwargs)
+        # assume text is never paginated
         return resp  # type: ignore
 
     # TODO: refactor this method so it looks like the other methods, while still supporting pagination.
@@ -2489,7 +2843,7 @@ class HarborAsyncClient:
         headers: Optional[Dict[str, Any]] = None,
         follow_links: bool = True,
         **kwargs: Any,
-    ) -> JSONType:
+    ) -> Tuple[JSONType, Optional[str]]:
         """Sends a GET request to the Harbor API.
         Returns JSON unless the response is text/plain.
 
@@ -2518,36 +2872,16 @@ class HarborAsyncClient:
         check_response_status(resp)
         j = handle_optional_json_response(resp)
         if j is None:
-            return resp.text  # type: ignore # FIXME: resolve this ASAP (use overload?)
+            return resp.text, None  # type: ignore # FIXME: resolve this ASAP (use overload?)
 
-        # If we have "Link" in headers, we need to handle paginated results
-        if (link := resp.headers.get("link")) and 'rel="next"' in link and follow_links:
+        # If we have "Link" in headers, we need to parse the next page link
+        if follow_links and (link := resp.headers.get("link")):
             logger.debug("Handling paginated results. Header value: {}", link)
-            j = await self._handle_pagination(j, link)  # recursion (refactor?)
-        return j
+            return j, parse_pagination_url(link)
 
-    async def _handle_pagination(self, data: JSONType, link: str) -> JSONType:
-        """Handles paginated results by recursing until all results are returned."""
-        # Parse the link header value and get next page URL
-        next_url = parse_pagination_url(link)
-
-        # ignoring params and only using the next URL
-        # the next URL should contain the original params with page number adjusted
-        j = await self.get(next_url)
-
-        if not isinstance(j, list) or not isinstance(data, list):
-            logger.warning(
-                "Unable to handle paginated results, received non-list value. URL: {}",
-                next_url,
-            )
-            # TODO: add more diagnostics info here
-            return data
-
-        data.extend(j)
-        return data
+        return j, None
 
     # NOTE: POST is not idempotent, should we still retry?
-    # TODO: fix abstraction of post/_post. Put everything into _post?
     @backoff.on_exception(backoff.expo, RequestError, max_tries=1)
     async def post(
         self,
